import MPI.*;

public class assign5
{
	static final int s_nPoints=10;
	static final int s_nTime_iterations=10;

	public static void main(String args[])
		throws java.io.IOException
	{
		/* initialize MPI */
		MPI.Init(args);
		nProcessors=MPI.COMM_WORLD.Size();
		idThis=MPI.COMM_WORLD.Rank();

		/* initialize storage arrays */
		initialize_arrays();

		/* initial calculations */
        tau=0.33*0.33;
		nCurrent_time=3;

		/* initial default neighbor values */
		dNeighbor_left[0]=0;
		dNeighbor_right[0]=0;

		/* identify left and right neighbors (NOTE: -1 means no neighbors (take values as 0) ) */
		idNeighbor_left=idThis-1;
		idNeighbor_right=idThis+1;
		if (idNeighbor_right>=nProcessors)
			idNeighbor_right=-1;
		
		/* calculate times 1 & 2 */
		generate_points(s_nPoints/nProcessors);
		calculate_equation();
		
		print_starting_values();

		while (nCurrent_time<s_nTime_iterations)
		{
			send_left();
			receive_right();
			send_right();
			receive_left();
	
			/* calculate equation */
			calculate_equation();

			print_answers();

			increment_time();
		}

        /* clean up */
		MPI.Finalize();
	}

	static double[] initialize_one_array()
	{
		return new double[s_nPoints/nProcessors];
	}

	static void initialize_arrays()
	{
		/* initializes memory for arrays, and set their values to 0 */
		arT_plus_1=initialize_one_array();
		arT=initialize_one_array();
		arT_minus_1=initialize_one_array();

        dNeighbor_left=new double[1];
	    dNeighbor_right=new double[1];
    }

	/* generates random points for Time Steps 1 and 2 for the processors x space */
	static void generate_points(int nPoints)
	{
		for (int i=0; i<nPoints; i++)
		{
			arT_minus_1[i]=Math.random();
			arT[i]=Math.random();
		}
	}

	static void receive_left()
	{
		if (idNeighbor_left!=-1)
		{
			MPI.COMM_WORLD.Recv(dNeighbor_left,0,1,MPI.DOUBLE,
			idNeighbor_left,nCurrent_time); //,statusLeft);
		}
		else
			dNeighbor_left[0]=0;
	}

	static void receive_right()
	{
		if (idNeighbor_right!=-1)
		{
			MPI.COMM_WORLD.Recv(dNeighbor_right,0,1,MPI.DOUBLE,
			idNeighbor_right,nCurrent_time);
		}
		else
			dNeighbor_right[0]=0;
	}
	
	static void send_left()
	{	
		/* send points to left neighbor */
		if (idNeighbor_left!=-1)
		{
			MPI.COMM_WORLD.Send(arT,0,1,
			    MPI.DOUBLE,idNeighbor_left,nCurrent_time);
		}
	}

	static void send_right()
	{
		/* send points to right neighbor */
		if (idNeighbor_right!=-1)
		{
			MPI.COMM_WORLD.Send(arT,(s_nPoints/nProcessors)-1,1,
			    MPI.DOUBLE,idNeighbor_right,nCurrent_time);
		}
	}

	static void calculate_equation()
	{
		int i;
		for (i=0; i<s_nPoints/nProcessors; i++)
		{
    		arT_plus_1[i]=(2*arT[i])-arT_minus_1[i];

		    /* left most case */
		    if (i==0)
    			arT_plus_1[i]+=tau*(dNeighbor_left[0]-
				    (2.0*arT[i])+arT[i+1]);

    		/* right most case */
		    else if (i==s_nPoints/nProcessors-1)
    			arT_plus_1[i]+=tau*(arT[i-1]-
				    (2.0*arT[i])+dNeighbor_right[0]);

    		/* generic case */
	 	    else
    			arT_plus_1[i]+=tau*(arT[i-1]-(2.0*arT[i])+arT[i+1]);		
		    }
	}

	static void print_answers()
	{
		int 		n,nPoints;
	
		nPoints=s_nPoints/nProcessors;
		
	 	/* print this processors answers */
        System.out.print("Time "+(nCurrent_time-1)+"  (Processor "+idThis+"): ");
		for (n=0; n<nPoints; n++)
            System.out.print(arT_plus_1[n]+" ");

        System.out.println();
	}

	static void print_starting_values()
	{
		int 		n,nPoints;
	
		nPoints=s_nPoints/nProcessors;
	
		/* print this processors answers */
        System.out.print("Time 0 (Processor "+idThis+"): ");
		for (n=0; n<nPoints; n++)
            System.out.print(arT_minus_1[n]+" ");

        System.out.println();
	
		/* print this processors answers */
        System.out.print("Time 1 (Processor "+idThis+"): ");
		for (n=0; n<nPoints; n++)
            System.out.print(arT[n]+" ");

        System.out.println();
	}
	
	static void increment_time()
	{
		/* Move T arrays left one (i.e., ar(T-1)=ar(T); ar(T)=ar(T+1)) */
	
		arT_minus_1=arT;
		arT=arT_plus_1;
		arT_plus_1=initialize_one_array();
	
		nCurrent_time++;
	}
	
	// Data members
	static double	arT_plus_1[],
				    arT_minus_1[],
				    arT[];
    static int		idThis,nProcessors,nCurrent_time;
	static double	tau;

	/* Global data to hold requested neighbors values */
	static int		idNeighbor_left,idNeighbor_right;
	static double	dNeighbor_left[];
	static double	dNeighbor_right[];
//	static MPI.Status	statusRight,statusLeft;
} // class assign5

